from onegov.core.orm import Base
from onegov.core.orm.mixins import ContentMixin
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UTCDateTime
from onegov.core.orm.types import UUID
from onegov.user import User
from onegov.user import UserGroup
from sedate import utcnow
from sqlalchemy import Column
from sqlalchemy import Enum
from sqlalchemy import ForeignKey
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import HSTORE
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.orm import backref
from sqlalchemy.orm import relationship
from uuid import uuid4


class OfficialNotice(Base, ContentMixin, TimestampMixin):

    """ Defines an official notice.

    The notice follows a typcical state transition: drafted by an editor ->
    submitted by and editor to a publisher -> accepted by a publisher ->
    published by the publisher. It may be alternatively rejected by a publisher
    when submitted.

    The notice typically has a title and a text, belongs to a user and a user
    group, appears in one ore more issues and belongs to one or more categories
    and organizations.

    You can set the date of the first issue besides setting the issues (to
    allow filtering by date for example).

    You can set the category and organization directly instead of using the
    HSTOREs (or use both).

    """

    __tablename__ = 'official_notices'

    #: the type of the item, this can be used to create custom polymorphic
    #: subclasses of this class. See
    #: `<http://docs.sqlalchemy.org/en/improve_toc/\
    #: orm/extensions/declarative/inheritance.html>`_.
    type = Column(Text, nullable=True)

    __mapper_args__ = {
        'polymorphic_on': type
    }

    #: The internal ID of the notice.
    id = Column(UUID, primary_key=True, default=uuid4)

    #: A nice ID usable for the url, readable by humans.
    name = Column(Text)

    #: The state of the notice.
    state = Column(
        Enum(
            'drafted',
            'submitted',
            'rejected',
            'imported',
            'accepted',
            'published',
            name='official_notice_state'
        ),
        nullable=False,
        default='drafted'
    )

    #: The title of the notice.
    title = Column(Text, nullable=False)

    #: The text of the notice.
    text = Column(Text, nullable=True)

    #: The author of the notice.
    author_name = Column(Text, nullable=True)

    #: The place (part of the signature).
    author_place = Column(Text, nullable=True)

    #: The date (part of the signature)
    author_date = Column(UTCDateTime, nullable=True)

    #: A note to the notice.
    note = Column(Text, nullable=True)

    #: The issues this notice appears in.
    _issues = Column(
        MutableDict.as_mutable(HSTORE), name='issues', nullable=True
    )

    @property
    def issues(self):
        return self._issues or {}

    @issues.setter
    def issues(self, value):
        if isinstance(value, dict):
            self._issues = value
        else:
            self._issues = {item: None for item in value}

    #: The date of the first issue of the notice.
    first_issue = Column(UTCDateTime, nullable=True)

    #: The expiry date of the notice
    expiry_date = Column(UTCDateTime, nullable=True)

    @property
    def expired(self):
        """ Returns True, if the notice is expired. """

        if self.expiry_date:
            return self.expiry_date < utcnow()
        return False

    #: The categories of this notice.
    _categories = Column(
        MutableDict.as_mutable(HSTORE), name='categories', nullable=True
    )

    @property
    def categories(self):
        return self._categories or {}

    @categories.setter
    def categories(self, value):
        if isinstance(value, dict):
            self._categories = value
        else:
            self._categories = {item: None for item in value}

    #: The category of the notice.
    category = Column(Text, nullable=True)

    #: The organization this notice belongs to.
    organization = Column(Text, nullable=True)

    #: The categories of this notice.
    _organizations = Column(
        MutableDict.as_mutable(HSTORE), name='organizations', nullable=True
    )

    @property
    def organizations(self):
        return self._organizations or {}

    @organizations.setter
    def organizations(self, value):
        if isinstance(value, dict):
            self._organizations = value
        else:
            self._organizations = {item: None for item in value}

    #: The user that owns this notice.
    user_id = Column(UUID, ForeignKey(User.id), nullable=True)
    user = relationship(
        User, backref=backref('official_notices', lazy='select')
    )

    #: The group that owns this notice.
    group_id = Column(UUID, ForeignKey(UserGroup.id), nullable=True)
    group = relationship(
        UserGroup, backref=backref('official_notices', lazy='select')
    )

    #: The source from where this notice has been imported.
    source = Column(Text, nullable=True)

    def submit(self):
        """ Submit a drafted notice. """

        assert self.state == 'drafted' or self.state == 'rejected'
        self.state = 'submitted'

    def reject(self):
        """ Reject a submitted notice. """

        assert self.state == 'submitted'
        self.state = 'rejected'

    def accept(self):
        """ Accept a submitted notice. """

        assert self.state == 'submitted' or self.state == 'imported'
        self.state = 'accepted'

    def publish(self):
        """ Publish an accepted notice. """

        assert self.state == 'accepted'
        self.state = 'published'



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.orm.types import UTCDateTime
from onegov.core.upgrade import upgrade_task
from sqlalchemy import Column
from sqlalchemy import Enum
from sqlalchemy import Text
from sqlalchemy.dialects.postgresql import HSTORE


@upgrade_task('Add categories column to official notices')
def add_categories(context):
    if not context.has_column('official_notices', 'categories'):
        context.operations.add_column(
            'official_notices',
            Column('categories', HSTORE, nullable=True)
        )


@upgrade_task('Add organizations column to official notices')
def add_organizations(context):
    if not context.has_column('official_notices', 'organizations'):
        context.operations.add_column(
            'official_notices',
            Column('organizations', HSTORE, nullable=True)
        )


@upgrade_task('Add author fields to official notices')
def add_author_fields(context):
    if not context.has_column('official_notices', 'author_name'):
        context.operations.add_column(
            'official_notices',
            Column('author_name', Text, nullable=True)
        )
    if not context.has_column('official_notices', 'author_place'):
        context.operations.add_column(
            'official_notices',
            Column('author_place', Text, nullable=True)
        )
    if not context.has_column('official_notices', 'author_date'):
        context.operations.add_column(
            'official_notices',
            Column('author_date', UTCDateTime, nullable=True)
        )


@upgrade_task('Add an imported state to official notices')
def add_imported_state_to_notices(context):
    old = ['drafted', 'submitted', 'published', 'rejected', 'accepted']
    new = old + ['imported']
    old_type = Enum(*old, name='official_notice_state')
    new_type = Enum(*new, name='official_notice_state')
    tmp_type = Enum(*new, name='_official_notice_state')

    tmp_type.create(context.operations.get_bind(), checkfirst=False)
    context.operations.execute(
        'ALTER TABLE official_notices ALTER COLUMN state '
        'TYPE _official_notice_state USING state::text::_official_notice_state'
    )
    old_type.drop(context.operations.get_bind(), checkfirst=False)
    new_type.create(context.operations.get_bind(), checkfirst=False)
    context.operations.execute(
        'ALTER TABLE official_notices ALTER COLUMN state '
        'TYPE official_notice_state USING state::text::official_notice_state'
    )
    tmp_type.drop(context.operations.get_bind(), checkfirst=False)


@upgrade_task('Add a souurce column to official notices')
def add_source_column_to_notices(context):
    if not context.has_column('official_notices', 'source'):
        context.operations.add_column(
            'official_notices',
            Column('source', Text, nullable=True)
        )


@upgrade_task('Add an expiry date column to official notices')
def add_expiry_date_column_to_notices(context):
    if not context.has_column('official_notices', 'expiry_date'):
        context.operations.add_column(
            'official_notices',
            Column('expiry_date', UTCDateTime, nullable=True)
        )


@upgrade_task('Add note column to official notices')
def add_note_to_notices(context):
    if not context.has_column('official_notices', 'note'):
        context.operations.add_column(
            'official_notices',
            Column('note', Text, nullable=True)
        )



from onegov.core.collection import Pagination
from onegov.core.utils import increment_name
from onegov.core.utils import normalize_for_url
from onegov.notice.models import OfficialNotice
from onegov.user import User
from onegov.user import UserGroup
from sqlalchemy import asc
from sqlalchemy import cast
from sqlalchemy import desc
from sqlalchemy import func
from sqlalchemy import inspect
from sqlalchemy import or_
from sqlalchemy import String


def get_unique_notice_name(name, session, model_class):
    """ Create a unique, URL-friendly name. """

    # it's possible for `normalize_for_url` to return an empty string...
    name = normalize_for_url(name) or "notice"

    while session.query(model_class.name).\
            filter(model_class.name == name).first():
        name = increment_name(name)

    return name


class OfficialNoticeCollectionPagination(Pagination):

    def __init__(
        self,
        session,
        page=0,
        state=None,
        term=None,
        order=None,
        direction=None,
        issues=None,
        categories=None,
        organizations=None,
        user_ids=None,
        group_ids=None
    ):
        self.session = session
        self.page = page
        self.state = state
        self.term = term
        self.order = order or 'first_issue'
        self.direction = direction or 'asc'
        self.issues = issues
        self.categories = categories
        self.organizations = organizations
        self.user_ids = user_ids or []
        self.group_ids = group_ids or []

    def __eq__(self, other):
        return (
            self.page == other.page
            and self.state == other.state
            and self.term == other.term
            and self.order == other.order
            and self.direction == other.direction
            and self.issues == other.issues
            and self.categories == other.categories
            and self.organizations == other.organizations
            and self.user_ids == other.user_ids
            and self.group_ids == other.group_ids
        )

    def subset(self):
        return self.query()

    @property
    def page_index(self):
        return self.page

    def page_by_index(self, index):
        return self.__class__(
            self.session,
            page=index,
            state=self.state,
            term=self.term,
            order=self.order,
            direction=self.direction,
            issues=self.issues,
            categories=self.categories,
            organizations=self.organizations,
            user_ids=self.user_ids,
            group_ids=self.group_ids
        )

    def for_state(self, state):
        """ Returns a new instance of the collection with the given state. """

        return self.__class__(
            self.session,
            state=state,
            term=self.term,
            order=self.order,
            direction=self.direction,
            issues=self.issues,
            categories=self.categories,
            organizations=self.organizations,
            user_ids=self.user_ids,
            group_ids=self.group_ids
        )

    def for_term(self, term):
        """ Returns a new instance of the collection with the given term. """

        return self.__class__(
            self.session,
            state=self.state,
            term=term,
            order=self.order,
            direction=self.direction,
            issues=self.issues,
            categories=self.categories,
            organizations=self.organizations,
            user_ids=self.user_ids,
            group_ids=self.group_ids
        )

    def for_order(self, order, direction=None):
        """ Returns a new instance of the collection with the given ordering.
        Inverts the direction if the new ordering is the same as the old one
        and an explicit ordering is not defined.

        """
        if direction is not None:
            descending = direction == 'desc'
        elif self.order != order:
            descending = False
        else:
            descending = self.direction != 'desc'

        return self.__class__(
            self.session,
            state=self.state,
            term=self.term,
            order=order,
            direction='desc' if descending else 'asc',
            issues=self.issues,
            categories=self.categories,
            organizations=self.organizations,
            user_ids=self.user_ids,
            group_ids=self.group_ids
        )

    def for_organizations(self, organizations):
        """ Returns a new instance of the collection with the given
        organizations.

        """

        return self.__class__(
            self.session,
            state=self.state,
            term=self.term,
            order=self.order,
            direction=self.direction,
            issues=self.issues,
            categories=self.categories,
            organizations=organizations,
            user_ids=self.user_ids,
            group_ids=self.group_ids
        )

    def for_categories(self, categories):
        """ Returns a new instance of the collection with the given categories.

        """

        return self.__class__(
            self.session,
            state=self.state,
            term=self.term,
            order=self.order,
            direction=self.direction,
            issues=self.issues,
            categories=categories,
            organizations=self.organizations,
            user_ids=self.user_ids,
            group_ids=self.group_ids
        )


class OfficialNoticeCollection(OfficialNoticeCollectionPagination):
    """ Manage a list of official notices. """

    @property
    def model_class(self):
        return OfficialNotice

    @property
    def term_columns(self):
        """ The columns used for full text search. """

        return [
            cast(self.model_class.id, String),
            self.model_class.title,
            self.model_class.text,
            self.model_class.author_name,
            self.model_class.author_place,
            self.model_class.category,
            self.model_class.organization,
            UserGroup.name,
            User.realname,
            User.username
        ]

    def filter_query(self, query):
        """ Filters the given query by the state of the collection. """

        if self.state:
            query = query.filter(self.model_class.state == self.state)
        if self.term:
            term = '%{}%'.format(self.term)
            query = query.filter(
                or_(
                    *[column.ilike(term) for column in self.term_columns]
                )
            )
        if self.user_ids:
            query = query.filter(self.model_class.user_id.in_(self.user_ids))
        if self.group_ids:
            query = query.filter(self.model_class.group_id.in_(self.group_ids))
        if self.issues:
            query = query.filter(self.model_class._issues.has_any(self.issues))
        if self.categories:
            query = query.filter(
                self.model_class._categories.has_any(self.categories)
            )
        if self.organizations:
            query = query.filter(
                self.model_class._organizations.has_any(self.organizations)
            )

        return query

    def order_query(self, query):
        """ Orders the given query by the state of the collection. """

        direction = desc if self.direction == 'desc' else asc
        if self.order in inspect(self.model_class).columns.keys():
            attribute = getattr(self.model_class, self.order)
        elif self.order == 'group.name':
            attribute = func.coalesce(UserGroup.name, '')
        elif self.order == 'user.realname':
            attribute = func.coalesce(User.realname, '')
        elif self.order == 'user.username':
            attribute = func.coalesce(User.username, '')
        elif self.order == 'user.name':
            attribute = func.coalesce(User.realname, User.username, '')
        else:
            attribute = self.model_class.first_issue

        return query.order_by(None).order_by(direction(attribute))

    def query(self):
        """ Returns a filtered and sorted query.

        Filters by:
        - notice.state matches state
        - notice.user_id is in user_ids
        - notice.group_id is in group_ids
        - notice.issues has any of the issues
        - term is in title, text, category, organization, groupname, usernames

        Orders by:
        - any notice columns
        - group name (group.name)
        - users real name (user.realname)
        - username (user.username)
        - username or users real name (user.name)

        """

        query = self.session.query(self.model_class)
        query = query.join(self.model_class.group, isouter=True)
        query = query.join(self.model_class.user, isouter=True)
        query = self.filter_query(query)
        query = self.order_query(query)
        return query

    def _get_unique_name(self, name):
        """ Create a unique, URL-friendly name. """

        return get_unique_notice_name(name, self.session, self.model_class)

    def add(self, title, text, **optional):
        """ Add a new notice.

        A unique, URL-friendly name is created automatically for this notice
        using the title and optionally numbers for duplicate names.

        Returns the created notice.
        """
        issues = optional.pop('issues', None)
        categories = optional.pop('categories', None)
        organizations = optional.pop('organizations', None)
        notice = self.model_class(
            state='drafted',
            name=self._get_unique_name(title),
            title=title,
            text=text,
            **optional
        )
        if issues:
            notice.issues = issues
        if categories:
            notice.categories = categories
        if organizations:
            notice.organizations = organizations

        self.session.add(notice)
        self.session.flush()

        return notice

    def delete(self, notice):
        """ Delete an notice. """

        self.session.delete(notice)
        self.session.flush()

    def by_name(self, name):
        """ Returns a notice by its URL-friendly name. """

        query = self.session.query(self.model_class)
        query = query.filter(self.model_class.name == name)
        return query.first()

    def by_id(self, id):
        """ Returns a notice by its id. """

        query = self.session.query(self.model_class)
        query = query.filter(self.model_class.id == id)
        return query.first()



from datetime import datetime
from datetime import timezone
from onegov.notice import OfficialNoticeCollection
from onegov.user import UserCollection
from onegov.user import UserGroupCollection
from transaction import commit


def test_notice_collection(session):
    notices = OfficialNoticeCollection(session)
    assert notices.query().count() == 0

    notice_1 = notices.add(
        title='Important Announcement',
        text='<em>Important</em> things happened!',
        category='important',
        organization='onegov'
    )
    notice_1.submit()
    notice_1.accept()
    notice_1.publish()

    commit()

    notice_1 = notices.query().one()
    assert notice_1.title == 'Important Announcement'
    assert notice_1.text == '<em>Important</em> things happened!'
    assert notice_1.category == 'important'
    assert notice_1.organization == 'onegov'

    notice_2 = notices.add(
        title='Important Announcement',
        text='<em>Important</em> things happened!'
    )
    notice_2.submit()

    assert notices.query().count() == 2
    assert notices.for_state('drafted').query().count() == 0
    assert notices.for_state('submitted').query().count() == 1
    assert notices.for_state('accepted').query().count() == 0
    assert notices.for_state('published').query().count() == 1

    assert notices.by_name('important-announcement')
    assert notices.by_name('important-announcement-1')

    assert notices.by_id(notice_1.id)
    assert notices.by_id(notice_2.id)

    notices.delete(notice_1)
    notices.delete(notice_2)

    assert notices.query().count() == 0
    assert notices.for_state('drafted').query().count() == 0
    assert notices.for_state('submitted').query().count() == 0
    assert notices.for_state('accepted').query().count() == 0
    assert notices.for_state('published').query().count() == 0


def test_notice_collection_search(session):
    groups = UserGroupCollection(session)
    group_a = groups.add(name='GroupA')
    group_b = groups.add(name='GroupB1')

    users = UserCollection(session)
    user_a = users.add('a@example.org', 'password', 'editor', realname='Hans')
    user_b = users.add('b@example.org', 'password', 'editor')

    notices = OfficialNoticeCollection(session)

    notice = notices.add(
        title='First',
        text='Lorem Ipsum',
        category='Cat1',
        group=group_a
    )
    notice.state = 'drafted'

    notice = notices.add(
        title='Second',
        text='A text',
        category='Cat1',
        organization='Org1',
        user=user_a,
        group=group_b
    )
    notice.state = 'submitted'

    notice = notices.add(
        title='Third',
        text='Anöther text',
        category='Cat2'
    )
    notice.state = 'drafted'

    notice = notices.add(
        title='Fourth',
        text='A fourth text',
        organization='Org2'
    )
    notice.state = 'published'

    notice = notices.add(
        title='Fifth',
        text='Lorem Ipsum',
        user=user_a
    )
    notice.state = 'rejected'

    notice = notices.add(
        title='Sixt',
        text='<p>Six</p>',
        author_name='Cysat',
        user=user_b,
        group=group_a
    )
    notice.state = 'published'

    notice = notices.add(
        title='Sübent',
        text='Sübent',
        author_place='Wynmärkt'
    )
    notice.state = 'drafted'

    assert notices.query().count() == 7

    assert notices.for_term('Third').query().count() == 1

    assert notices.for_term('ourth').query().count() == 1

    assert notices.for_term('ipsum').query().count() == 2
    assert notices.for_term('ipsum').for_state('rejected').query().count() == 1

    assert notices.for_term('six').query().count() == 1
    assert notices.for_term('six').for_state('rejected').query().count() == 0

    assert notices.for_term('üb').query().count() == 1

    assert notices.for_term('Cat1').query().count() == 2
    assert notices.for_term('Cat2').query().count() == 1
    assert notices.for_term('Cat').query().count() == 3

    assert notices.for_term('Org1').query().count() == 1
    assert notices.for_term('Org').query().count() == 4

    assert notices.for_term('@example.org').query().count() == 3
    assert notices.for_term('ans').query().count() == 2

    assert notices.for_term('group').query().count() == 3
    assert notices.for_term('groupb').query().count() == 1

    assert notices.for_term('Cysat').query().count() == 1

    assert notices.for_term('Wynmärkt').query().count() == 1

    assert notices.for_term(str(notice.id).split('-')[0]).query().count() == 1


def test_notice_collection_users_and_groups(session):
    groups = UserGroupCollection(session)
    group_ab = groups.add(name='AB')
    group_c = groups.add(name='C')
    group_d = groups.add(name='D')

    users = UserCollection(session)
    user_a = users.add('a@example.org', 'password', 'editor', group=group_ab)
    user_b = users.add('b@example.org', 'password', 'editor', group=group_ab)
    user_c = users.add('c@example.org', 'password', 'editor', group=group_c)
    user_d = users.add('d@example.org', 'password', 'editor', group=group_d)
    user_e = users.add('e@example.org', 'password', 'editor')

    notices = OfficialNoticeCollection(session)
    for title, user in (
        ('A1', user_a),
        ('A2', user_a),
        ('A3', user_a),
        ('B1', user_b),
        ('B2', user_b),
        ('C1', user_c),
        ('C2', user_c),
        ('D1', user_d),
        ('D2', user_d),
        ('D3', user_d),
    ):
        notices.add(
            title=title,
            text='text',
            user_id=user.id,
            group_id=user.group.id if user.group else None
        )

    assert notices.query().count() == 10

    # test users
    notices.user_ids = [user_a.id]
    assert sorted([n.title for n in notices.query()]) == ['A1', 'A2', 'A3']

    notices.user_ids = [user_b.id]
    assert sorted([n.title for n in notices.query()]) == ['B1', 'B2']

    notices.user_ids = [user_c.id]
    assert sorted([n.title for n in notices.query()]) == ['C1', 'C2']

    notices.user_ids = [user_d.id]
    assert sorted([n.title for n in notices.query()]) == ['D1', 'D2', 'D3']

    notices.user_ids = [user_e.id]
    assert sorted([n.title for n in notices.query()]) == []

    notices.user_ids = [user_b.id, user_c.id]
    assert sorted([n.title for n in notices.query()]) == [
        'B1', 'B2', 'C1', 'C2'
    ]

    # test groups
    notices.user_ids = []

    notices.group_ids = [group_ab.id]
    assert sorted([n.title for n in notices.query()]) == [
        'A1', 'A2', 'A3', 'B1', 'B2'
    ]

    notices.group_ids = [group_c.id]
    assert sorted([n.title for n in notices.query()]) == ['C1', 'C2']

    notices.group_ids = [group_d.id]
    assert sorted([n.title for n in notices.query()]) == ['D1', 'D2', 'D3']

    notices.group_ids = [group_ab.id, group_d.id]
    assert sorted([n.title for n in notices.query()]) == [
        'A1', 'A2', 'A3', 'B1', 'B2', 'D1', 'D2', 'D3'
    ]

    # test users and groups
    notices.group_ids = [group_ab.id, group_d.id]
    notices.user_ids = [user_b.id, user_d.id]
    assert sorted([n.title for n in notices.query()]) == [
        'B1', 'B2', 'D1', 'D2', 'D3'
    ]


def test_notice_collection_issues(session):
    notices = OfficialNoticeCollection(session)
    notices.add(title='a', text='text', issues=None)
    notices.add(title='b', text='text', issues=[])
    notices.add(title='c', text='text', issues=['1', '2'])
    notices.add(title='d', text='text', issues=['1'])
    notices.add(title='e', text='text', issues=['2'])
    notices.add(title='f', text='text', issues=['2', '3'])
    notices.add(title='g', text='text', issues=['4'])

    for issues, result in (
        (None, ['a', 'b', 'c', 'd', 'e', 'f', 'g']),
        ([], ['a', 'b', 'c', 'd', 'e', 'f', 'g']),
        (['1'], ['c', 'd']),
        (['2'], ['c', 'e', 'f']),
        (['3'], ['f']),
        (['4'], ['g']),
        (['1', '4'], ['c', 'd', 'g']),
        (['2', '3'], ['c', 'e', 'f']),
    ):
        notices.issues = issues
        assert sorted([notice.title for notice in notices.query()]) == result


def test_notice_collection_categories(session):
    notices = OfficialNoticeCollection(session)
    notices.add(title='a', text='text', categories=None)
    notices.add(title='b', text='text', categories=[])
    notices.add(title='c', text='text', categories=['1', '2'])
    notices.add(title='d', text='text', categories=['1'])
    notices.add(title='e', text='text', categories=['2'])
    notices.add(title='f', text='text', categories=['2', '3'])
    notices.add(title='g', text='text', categories=['4'])

    for categories, result in (
        (None, ['a', 'b', 'c', 'd', 'e', 'f', 'g']),
        ([], ['a', 'b', 'c', 'd', 'e', 'f', 'g']),
        (['1'], ['c', 'd']),
        (['2'], ['c', 'e', 'f']),
        (['3'], ['f']),
        (['4'], ['g']),
        (['1', '4'], ['c', 'd', 'g']),
        (['2', '3'], ['c', 'e', 'f']),
    ):
        assert sorted([
            n.title for n in notices.for_categories(categories).query()
        ]) == result


def test_notice_collection_organizations(session):
    notices = OfficialNoticeCollection(session)
    notices.add(title='a', text='text', organizations=None)
    notices.add(title='b', text='text', organizations=[])
    notices.add(title='c', text='text', organizations=['1', '2'])
    notices.add(title='d', text='text', organizations=['1'])
    notices.add(title='e', text='text', organizations=['2'])
    notices.add(title='f', text='text', organizations=['2', '3'])
    notices.add(title='g', text='text', organizations=['4'])

    for organizations, result in (
        (None, ['a', 'b', 'c', 'd', 'e', 'f', 'g']),
        ([], ['a', 'b', 'c', 'd', 'e', 'f', 'g']),
        (['1'], ['c', 'd']),
        (['2'], ['c', 'e', 'f']),
        (['3'], ['f']),
        (['4'], ['g']),
        (['1', '4'], ['c', 'd', 'g']),
        (['2', '3'], ['c', 'e', 'f']),
    ):
        assert sorted([
            n.title for n in notices.for_organizations(organizations).query()
        ]) == result


def test_notice_collection_order(session):
    groups = UserGroupCollection(session)
    group_c = groups.add(name='C').id
    group_d = groups.add(name='D').id

    users = UserCollection(session)
    user_a = users.add('a@example.org', 'password', 'editor', realname='o').id
    user_b = users.add('b@example.org', 'password', 'editor', realname='p').id
    user_c = users.add('c@example.org', 'password', 'editor').id

    date_1 = datetime(2007, 10, 10, tzinfo=timezone.utc)
    date_2 = datetime(2007, 11, 11, tzinfo=timezone.utc)
    date_3 = datetime(2007, 12, 12, tzinfo=timezone.utc)

    notices = OfficialNoticeCollection(session)
    for title, text, user, group, organization, category, first_issue in (
        ('A', 'g', user_a, group_c, 'p', 'X', date_1),
        ('B', 'g', user_b, group_d, 'q', 'X', date_1),
        ('B', 'h', None, None, 'p', 'X', date_2),
        ('C', 'h', user_a, None, 'q', 'X', date_1),
        ('D', 'i', user_b, group_d, 'p', 'Y', date_1),
        ('E', 'j', user_c, group_c, 'q', 'Y', date_3),
    ):
        notices.add(
            title=title,
            text=text,
            user_id=user,
            group_id=group,
            organization=organization,
            category=category,
            first_issue=first_issue,
        )

    # Default ordering by issue date
    result = [n.first_issue for n in notices.query()]
    assert result == [date_1, date_1, date_1, date_1, date_2, date_3]

    # Explicit direction
    result = [n.title for n in notices.for_order('title', 'asc').query()]
    assert result == ['A', 'B', 'B', 'C', 'D', 'E']

    result = [n.title for n in notices.for_order('title', 'desc').query()]
    assert result == ['E', 'D', 'C', 'B', 'B', 'A']

    result = [n.title for n in notices.for_order('title', '').query()]
    assert result == ['A', 'B', 'B', 'C', 'D', 'E']

    result = [n.title for n in notices.for_order('title', 'xxx').query()]
    assert result == ['A', 'B', 'B', 'C', 'D', 'E']

    # Default direction
    # ... default
    result = [
        n.title
        for n in notices.for_order('text', 'desc').for_order('title').query()
    ]
    assert result == ['A', 'B', 'B', 'C', 'D', 'E']

    # ... flip direction
    result = [
        n.title
        for n in notices.for_order('title', 'asc').for_order('title').query()
    ]
    assert result == ['E', 'D', 'C', 'B', 'B', 'A']

    # Invalid
    result = [n.first_issue for n in notices.for_order('result').query()]
    assert result == [date_1, date_1, date_1, date_1, date_2, date_3]

    result = [n.first_issue for n in notices.for_order('users').query()]
    assert result == [date_1, date_1, date_1, date_1, date_2, date_3]

    result = [n.first_issue for n in notices.for_order(None).query()]
    assert result == [date_1, date_1, date_1, date_1, date_2, date_3]

    # Valid
    result = [n.text for n in notices.for_order('text').query()]
    assert result == ['g', 'g', 'h', 'h', 'i', 'j']

    # ... user_id
    result = [n.user_id for n in notices.for_order('user_id').query()]
    assert result == sorted(2 * [user_a] + 2 * [user_b] + [user_c]) + [None]

    # ... organization
    result = [
        n.organization for n in notices.for_order('organization').query()
    ]
    assert result == ['p', 'p', 'p', 'q', 'q', 'q']

    # ... category
    result = [n.category for n in notices.for_order('category').query()]
    assert result == ['X', 'X', 'X', 'X', 'Y', 'Y']

    # ... name
    result = [n.name for n in notices.for_order('name').query()]
    assert result == ['a', 'b', 'b-1', 'c', 'd', 'e']

    # ... group.name
    result = [
        n.group.name if n.group else None
        for n in notices.for_order('group.name').query()
    ]
    assert result == [None, None, 'C', 'C', 'D', 'D']

    # ... user.name(s)
    result = [
        n.user.realname if n.user else None
        for n in notices.for_order('user.realname').query()
    ]
    assert result == [None, None, 'o', 'o', 'p', 'p']

    result = [
        n.user.username if n.user else None
        for n in notices.for_order('user.username').query()
    ]
    assert result == [
        None, 'a@example.org', 'a@example.org',
        'b@example.org', 'b@example.org', 'c@example.org'
    ]

    result = [
        n.user.realname or n.user.username if n.user else None
        for n in notices.for_order('user.name').query()
    ]
    assert result == [None, 'c@example.org', 'o', 'o', 'p', 'p']


def test_notice_collection_pagination(session):
    notices = OfficialNoticeCollection(session)

    assert notices.page_index == 0
    assert notices.pages_count == 0
    assert notices.batch == []

    for year in range(2008, 2013):
        for month in range(1, 13):
            notice = notices.add(
                title='Notice {0}-{1}'.format(year, month),
                text='A' if month % 3 else 'B'
            )
            if 2009 <= year:
                notice.submit()
            if 2010 <= year <= 2011:
                notice.accept()
            if 2011 <= year <= 2011:
                notice.publish()
            if 2012 <= year:
                notice.reject()

    assert notices.query().count() == 60

    drafted = notices.for_state('drafted')
    assert drafted.subset_count == 12
    assert len(drafted.next.batch) == 12 - drafted.batch_size

    submitted = notices.for_state('submitted')
    assert submitted.subset_count == 12
    assert len(submitted.next.batch) == 12 - submitted.batch_size

    accepted = notices.for_state('accepted')
    assert accepted.subset_count == 12
    assert len(accepted.next.batch) == 12 - accepted.batch_size

    published = notices.for_state('published')
    assert published.subset_count == 12
    assert len(published.next.batch) == 12 - published.batch_size

    rejected = notices.for_state('rejected')
    assert rejected.subset_count == 12
    assert len(rejected.next.batch) == 12 - rejected.batch_size
